* =====================================================================
*     P_{qq}^{PS}
! ---------------------------------------------------------------------
*     
*     Contribution from approximate four-loop splitting function with
*     variational parameter (-0.7 < aPS < 0.0).
*     This is the regular part of P3PS.
*
*     J. M. - 11/07/2022
*
      FUNCTION Pqqps3A(NN, a)
c      implicit none
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      double complex NN,Pqqps3A
      external Pqqps3exactA,ffsg1,ffsg2,ffsg3,ffsg4,ffsg14
      double complex Pqqps3exactA,ffsg1,ffsg2,ffsg3,ffsg4,ffsg14
      INTEGER n_dimension
      DOUBLE PRECISION N(4), C(4),  
     &        N_exact(4), N_a(4)
      DOUBLE PRECISION inv_A(4, 4)
      DATA PI,PI2/3.14159,9.8696/
*
      n_dimension=4
* 
*     Moments for the known exact small-x contribution (Vogt)
      N_exact(1) = -1.28827235D-01
      N_exact(2) = -2.42108771D-03
      N_exact(3) = -6.81270332D-05
      N_exact(4) = 4.79640375e-05
* 
*     Moments for the varied function multiplied by a (ln(1/x)/x)*(1-x)
      N_a(1) = 0.75D0
      N_a(2) = 0.04861111111078488D0
      N_a(3) = 0.012222222222363594D0
      N_a(4) = 0.004783163265306247D0
*
      DO i=1, n_dimension
        N(i) = N_exact(i) - a * N_a(i)
      ENDDO
*
      inv_A = reshape((/-4.33423515D+01, 1.04394869D+03, 
     &                  -3.26481918D+03, 2.49215628D+03, 1.01800794D+02, 
     &                  -2.29059693D+03, 7.08126548D+03, 
     &                  -5.37933607D+03, 1.28895321D+02, 
     &                  -3.28605015D+03, 1.09724714D+04, 
     &                  -8.75612151D+03, -9.56650450D0, 2.57515782D+02, 
     &                  -9.20800163D+02, 7.84949427D+02/), 
     &                 shape=(/4,4/))

* 
*     Matrix multiplication of inv_A and N
*
      DO i=1,n_dimension
        C(i) = 0.D0
        DO j=1,n_dimension
          C(i) = C(i) + inv_A(j, i) * N(j)
        ENDDO
      ENDDO
*
c      Pqqps3A = ((1.-X) * (C(1) * (1./X) + C(2) * (LOG(X)**2.) 
c     &         + C(3) * (X**2.) + C(4) * (LOG(1.-X)**2.) 
c     &         + a * LOG(1./X)/X) + Pqqps3exactA(X)) * (4*PI)**4.

      Pqqps3A = (C(1)*ffsg1(NN) + C(2)*ffsg2(NN)
     &         + C(3)*ffsg3(NN) + C(4)*ffsg4(NN)
     &     + a * ffsg14(NN) + Pqqps3exactA(NN)) * (4*PI)**4.
*     
      RETURN
      END
*
* ------------------------------------------------------------------------
*
*     This is the regular exact part of P3PS.
*
*     J. M. - 11/07/2022
*
      FUNCTION Pqqps3exactA(NN)
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      double complex NN,Pqqps3exactA
      external ffsg15
      double complex ffsg15
      DATA PI,PI2/3.14159,9.8696/
*
      CA = 3.0D0
      CF = 4.0D0/3.0D0
      AL1 = LOG(1.-X)
*
c      Pqqps3exactA = (((CA ** 2.0)*CF / (6.0D0 * (PI**4.))) * (
c     &                (82.0D0/81.0D0) + 2.0 * 1.20206D0) 
c     &                * (log(1.0 / x)**2.) / x)

      Pqqps3exactA = (((CA ** 2.0)*CF / (6.0D0 * (PI**4.))) * (
     &                (82.0D0/81.0D0) + 2.0 * 1.20206D0) 
     &                * ffsg15(NN))
*     
      RETURN
      END
*
* =====================================================================
*     P_{qg}
! ---------------------------------------------------------------------
*
*     Contribution from approximate four-loop splitting function with
*     variational parameter (-2.5 < aQG < -0.8).
*     This is the regular part of P3QG.
*
*     J. M. - 11/07/2022
*
      FUNCTION Pqg3A(NN, a)
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      double complex NN,Pqg3A
      external Pqg3exactA,ffsg8,ffsg10,ffsg11,ffsg12,ffsg14
      double complex Pqg3exactA,ffsg8,ffsg10,ffsg11,ffsg12,ffsg14
      INTEGER n_dimension
      DOUBLE PRECISION N(4), C(4), N_exact(4), 
     &        N_a(4)
      DOUBLE PRECISION inv_A(4, 4)
      DATA PI,PI2/3.14159,9.8696/
*
      n_dimension=4
* 
*     Moments for the known exact small-x contribution (Vogt)
      N_exact(1) = -0.32821791D0
      N_exact(2) = -0.0079068D0
      N_exact(3) = 0.00668085D0
      N_exact(4) = 0.0112976D0
* 
*     Moments for the varied function multiplied by a (ln(1/x)/x)
      N_a(1) = 0.9999999999999999D0
      N_a(2) = 0.11111111111076043D0
      N_a(3) = 0.04000000000013782D0
      N_a(4) = 0.02040816D0
*
      DO i=1, n_dimension
        N(i) = N_exact(i) - a * N_a(i)
      ENDDO
*
      inv_A = reshape((/1.50859345D1, -1.73502577D2,
     &     3.81181701D2, -2.25021001D2, 
     &     1.39350535D1, -2.33166358D2,
     &     5.53133581D2, -3.37644014D2, 
     &     3.92339377D0, -8.52278788D1,
     &     2.57493287D2, -1.79628628D2, 
     &     -1.15513419D-2, 2.74882781D-1,
     &     -9.19800075D-1, 7.20490480D-1/), 
     &                  shape=(/4,4/))
*
*     Matrix multiplication of inv_A and N
*
      DO i=1,n_dimension
        C(i) = 0.D0
        DO j=1,n_dimension
          C(i) = C(i) + inv_A(j, i) * N(j)
        ENDDO
      ENDDO
*
c      Pqg3A = (C(1) * LOG(X)**2. + C(2) * LOG(X) + C(3) * (X**2.) 
c     &          + C(4) * LOG(1.-X)**4.
c     &          + a * LOG(1./X)/X + Pqg3exactA(X)) * (4*PI)**4.

      Pqg3A = (C(1) * ffsg12(NN) + C(2) * ffsg11(NN) + C(3) * ffsg10(NN)
     &          + C(4) * ffsg8(NN)
     &          + a * ffsg14(NN) + Pqg3exactA(NN)) * (4*PI)**4.
*     
      RETURN
      END
*
* ---------------------------------------------------------------------------
*
*     This is the regular exact part of P3QG.
*
*     J. M. - 11/07/2022
*
      FUNCTION Pqg3exactA(NN)
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      double complex NN,Pqg3exactA
      external ffsg15
      double complex ffsg15
*
      DATA PI,PI2/3.14159,9.8696/
*
      CA = 3.0D0
      CF = 4.0D0/3.0D0
      AL1 = LOG(1.-X)
*
c      Pqg3exactA = ((CA ** 3.0)/ (6.0 * (PI**4.))) * 
c     &               ((82.0D0 / 81.0D0) + 2.0 * 1.20206D0) * 
c     &               (log(1.0 / x)**2.) / x
      Pqg3exactA = ((CA ** 3.0)/ (6.0 * (PI**4.))) * 
     &               ((82.0D0 / 81.0D0) + 2.0 * 1.20206D0) * 
     &               ffsg15(NN)
*
      RETURN
      END
*
* =====================================================================
*     P_{gq}
! ---------------------------------------------------------------------
*
*     Contribution from approximate four-loop splitting function with
*     variational parameter (-1.8 < aGQ < -1.5).
*     This is the regular part of P3GQ.
*
*     J. M. - 11/07/2022
*
      FUNCTION Pgq3A(NN, a)
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      double complex NN,Pgq3A
      external Pgq3exactA,ffsg6,ffsg9,ffsg13,ffsg14,ffsg15
      double complex Pgq3exactA,ffsg6,ffsg9,ffsg13,ffsg14,ffsg15
      INTEGER n_dimension
      DOUBLE PRECISION N(4), C(4), N_exact(4), 
     &        N_a(4)
      DOUBLE PRECISION inv_A(4, 4)
      DATA PI,PI2/3.14159,9.8696/
*
      n_dimension=4
*
* 
*     Moments for the known exact small-x contribution (Vogt)
      N_exact(1) = -0.78602265D0
      N_exact(2) = 0.05776039D0
      N_exact(3) = 0.04868481D0
      N_exact(4) = 0.04071129D0
* 
*     Moments for the varied function multiplied by a (ln^2(1/x)/x)
      N_a(1) = 1.9999999999999991D0
      N_a(2) = 0.07407407407348075D0
      N_a(3) = 0.016000000001040217D0
      N_a(4) = 0.0058309037900882415D0
*
      DO i=1, n_dimension
        N(i) = N_exact(i) - a * N_a(i)
      ENDDO
*
      inv_A = reshape((/-15.48054909D0, -39.22467343D0, 13.63767859D0, 
     &                 3.69946174D0, 442.68554891D0, 1038.70260338D0, 
     &                 -415.55730182D0,-113.79603728D0,-1263.91174116D0, 
     &                 -2938.1594667D0, 1302.32965951D0,362.67726047D0, 
     &                 874.63704072D0, 2025.64293423D0,-958.33374045D0, 
     &                 -272.56485302D0/), shape=(/4,4/))
*
*     Matrix multiplication of inv_A and N
      DO i=1,n_dimension
        C(i) = 0.D0
        DO j=1,n_dimension
          C(i) = C(i) + inv_A(i, j) * N(j)
        ENDDO
      ENDDO
*
c      Pgq3A = (C(1) * LOG(1./X)/X + C(2) * (LOG(X)**3.) + C(3) * X 
c     &          + C(4) * LOG(1.-X) + a * (LOG(1./X)**2.)/X 
c     &          + Pgq3exactA(X)) * (4*PI)**4.
      Pgq3A = (C(1) * ffsg14(NN) + C(2) * ffsg13(NN) + C(3) * ffsg9(NN) 
     &          + C(4) * ffsg6(NN) + a * ffsg15(NN)
     &          + Pgq3exactA(NN)) * (4*PI)**4.
*
      RETURN
      END
*
* ---------------------------------------------------------------------------
*
*     This is the regular exact part of P3GQ.
*
*     J. M. - 11/07/2022
*
      FUNCTION Pgq3exactA(NN)
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      double complex NN,Pgq3exactA
      external ffsg16
      double complex ffsg16
      DATA PI,PI2/3.14159,9.8696/
*
      CA = 3.0D0
      CF = 4.0D0/3.0D0
      AL1 = LOG(1.-X)
*
c      Pgq3exactA = ((CA ** 3.0) * CF / (3.0 * (PI**4.))) * 1.20206D0 
c     &               * (log(1.0 / x)**3.) / x
      Pgq3exactA = ((CA ** 3.0) * CF / (3.0 * (PI**4.))) * 1.20206D0 
     &               * ffsg16(NN)
*
      RETURN
      END
*
* =====================================================================
*     P_{gg}
! ---------------------------------------------------------------------
*
*     Contribution from approximate four-loop splitting function with
*     variational parameter (-5 < aGG < -15).
*     This is the regular part of P3GG.
*
*     J. M. - 11/07/2022
*
      FUNCTION Pgg3A(NN, a)
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      double complex NN,Pgg3A
      external Pgg3exactA,ffsg7,ffsg10,ffsg11,ffsg12,ffsg14
      double complex Pgg3exactA,ffsg7,ffsg10,ffsg11,ffsg12,ffsg14
      INTEGER n_dimension
      DOUBLE PRECISION N(4), C(4), N_exact(4), 
     &        N_a(4), inv_A(4, 4)
      DATA PI,PI2/3.14159,9.8696/
*
      n_dimension=4
* 
*     Moments for the known exact small-x contribution (Vogt)
      N_exact(1) = 6.94542399D0
      N_exact(2) = 0.01255781D0
      N_exact(3) = -0.24160401D0
      N_exact(4) = -0.27416992D0
* 
*     Moments for the varied function multiplied by a (ln(1/x)/x)
      N_a(1) = 0.9999999999999999D0
      N_a(2) = 0.11111111111076043D0
      N_a(3) = 0.04000000000013782D0
      N_a(4) = 0.02040816D0
*
      DO i=1, n_dimension
        N(i) = N_exact(i) - a * N_a(i)
      ENDDO
*
      inv_A = reshape((/1.57592805D1, -1.89525930D2,
     &     4.34798303D2, -2.67019531D2, 
     &     1.51046749D1, -2.60999383D2,
     &     6.46267166D2, -4.10596673D2, 
     &     6.05043429D0, -1.35844233D2,
     &     4.26863395D2, -3.12298274D2, 
     &     -3.86434270D-1, 9.19582568D0,
     &     -3.07706474D1, 2.41030189D1/), 
     &                  shape=(/4,4/))
*
*     Matrix multiplication of inv_A and N     
      DO i=1,n_dimension
        C(i) = 0.D0
        DO j=1,n_dimension
          C(i) = C(i) + inv_A(j, i) * N(j)
        ENDDO
      ENDDO
*
c      Pgg3A = (C(1) * LOG(X)**2. + C(2) * LOG(X) + C(3) * (X**2.) 
c     &          + C(4) * (LOG(1.-X)**2.) 
c     &          + a * LOG(1./X)/X + Pgg3exactA(X)) * (4*PI)**4.
      Pgg3A = (C(1) * ffsg12(NN) + C(2) * ffsg11(NN) + C(3) * ffsg10(NN)
     &          + C(4) * ffsg7(NN)
     &          + a * ffsg14(NN) + Pgg3exactA(NN)) * (4*PI)**4.
*
      RETURN
      END
*
* -----------------------------------------------------------------------------
*
*     This is the regular exact part of P3GG.
*
*     J. M. - 11/07/2022
*
      FUNCTION Pgg3exactA(NN)
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      double complex NN,Pgg3exactA
      external ffsg15,ffsg16
      double complex ffsg15,ffsg16
      DATA PI,PI2/3.14159,9.8696/
      DOUBLE PRECISION NF4
*
      CA = 3.0D0
      CF = 4.0D0/3.0D0
      AL1 = LOG(1.-X)
      NF4 = 4.0D0 ! Set Nf = 4 since the moments in the approximation 
                  ! are for 4 flavours.
*
      Pgg3exactA = ((CA ** 4.0D0) / (3.0D0 * (PI**4.0D0))) * 1.20206D0 
     &             * ffsg16(NN) + (1.0D0 / ((PI**4.0D0))) 
     &             * ((CA**4.0D0)*(-(1205.0D0/162.0D0)+(67.0D0/36.0D0) 
     &             * ((PI**2.0)/6.0D0)+(1.0D0/4.0D0) * (((PI**2.0) 
     &             / 6.0D0)**2.0)-(11.0D0/2.0D0)*1.20206D0) + NF4 
     &             * (CA ** 3.0) * (-(233.0D0/162.0D0)+(13.0D0/36.0D0) 
     &             * ((PI**2.0)/6.0D0)-(1.0D0/3.0D0) * 1.20206D0) + NF4 
     &             * (CA ** 2.0) * CF * ((617.0D0/243.0D0) 
     &             - (13.0D0/18.0D0)*((PI**2.0)/6.0D0) + (2.0D0/3.0D0) 
     &             * 1.20206D0))*(1.0D0/2.0D0)*(ffsg15(NN))
*
      RETURN
      END
*
